Goblin.IOError exception (Goblin v0.11.0)

Copy Markdown View Source

Raised when Goblin cannot read or write its underlying storage files.

The exception has the following fields:

  • :operation which describes which operation failed
  • :path which states which file it failed to operate on
  • :reason which describes why the operation failed