Raised when Goblin cannot read or write its underlying storage files.
The exception has the following fields:
:operationwhich describes which operation failed:pathwhich states which file it failed to operate on:reasonwhich describes why the operation failed